﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  /// <summary>
  /// Represents a weight calculator that gives every classification solution a weight based on the AdaBoost algorithm.
  /// </summary>
  [StorableClass]
  [Item("AdaBoostWeightCalculator", "Represents a weight calculator that gives every classification solution a weight based on the AdaBoost algorithm.")]
  public class AdaBoostWeightCalculator : WeightCalculator {

    public AdaBoostWeightCalculator()
      : base() {
    }

    [StorableConstructor]
    protected AdaBoostWeightCalculator(bool deserializing) : base(deserializing) { }
    protected AdaBoostWeightCalculator(AdaBoostWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AdaBoostWeightCalculator(this, cloner);
    }

    public override IEnumerable<double> CalculateWeights(ItemCollection<IClassificationSolution> classificationSolutions) {
      if (classificationSolutions.Count <= 0)
        return new List<double>();
      IClassificationProblemData problemData = classificationSolutions.ElementAt(0).ProblemData;
      double normalize = problemData.TrainingIndizes.Count() + problemData.TestIndizes.Count();
      List<double> distribution = Enumerable.Repeat(1.0 / normalize, Convert.ToInt32(normalize)).ToList();
      IEnumerable<double> targetValues = problemData.Dataset.GetDoubleValues(problemData.TargetVariable);

      List<double> weights = new List<double>();
      foreach (var solution in classificationSolutions) {
        double error = 0.0;
        //calculate the error

        //TODO: only works if training and test partition are not overlapping (has to be changed)
        IEnumerable<double> training = targetValues.Skip(problemData.TrainingPartition.Start).Take(problemData.TrainingIndizes.Count());
        IEnumerable<double> test = targetValues.Skip(problemData.TestPartition.Start).Take(problemData.TestIndizes.Count());

        IList<double> original = training.Concat(test).ToList();
        IList<double> estimated = solution.EstimatedTrainingClassValues.Concat(solution.EstimatedTestClassValues).ToList();

        for (int i = 0; i < original.Count; i++) {
          if (!original[i].Equals(estimated[i]))
            error += distribution[i];
        }
        //if the error is too big, the solution is not taken into account (weight would be negative)
        if (error > 0.5) {
          weights.Add(0);
          continue;
        }
        double beta = error / (1 - error);
        weights.Add(Math.Log(1 / beta));
        //Update the distribution
        normalize = 0.0;
        for (int i = 0; i < original.Count; i++) {
          if (original[i].Equals(estimated[i]))
            distribution[i] *= beta;
          normalize += distribution[i];
        }
        distribution = distribution.Select(x => x / normalize).ToList();
      }
      return weights;
    }
  }
}
